#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mygrid.hh"

template <class Grid> std::shared_ptr<Grid> constructGrid() {
  Dune::GridFactory<Grid> gridFactory;

#if DIM == 3
  Dune::FieldMatrix<double, 6, DIM> vertices;
#else
  Dune::FieldMatrix<double, 3, DIM> vertices;
#endif
  for (size_t i = 0; i < 2; ++i) {
    vertices[0][i] = MyGeometry::A[i];
    vertices[1][i] = MyGeometry::B[i];
    vertices[2][i] = MyGeometry::C[i];
#if DIM == 3
    vertices[3][i] = MyGeometry::A[i];
    vertices[4][i] = MyGeometry::B[i];
    vertices[5][i] = MyGeometry::C[i];
#endif
  }

#if DIM == 3
  for (size_t k = 0; k < 3; ++k) {
    vertices[k][2] = -MyGeometry::depth / 2.0;
    vertices[k + 3][2] = MyGeometry::depth / 2.0;
  }
#endif

  for (size_t i = 0; i < vertices.N(); ++i)
    gridFactory.insertVertex(vertices[i]);

  Dune::GeometryType cell;
#if DIM == 3
  cell.makeTetrahedron();
#else
  cell.makeTriangle();
#endif

#if DIM == 3
  std::vector<std::vector<unsigned int>> simplices = { { 0, 1, 2, 3 },
                                                       { 1, 2, 3, 4 },
                                                       { 2, 3, 4, 5 } };
#else
  std::vector<std::vector<unsigned int>> simplices = { { 0, 1, 2 } };
#endif

  // sanity-check choices of simplices
  for (size_t i = 0; i < simplices.size(); ++i) {
    Dune::FieldMatrix<double, DIM, DIM> check;
    for (size_t j = 0; j < DIM; ++j)
      check[j] = vertices[simplices[i][j + 1]] - vertices[simplices[i][j]];
    assert(check.determinant() > 0);
    gridFactory.insertElement(cell, simplices[i]);
  }
  return std::shared_ptr<Grid>(gridFactory.createGrid());
};

template <class GridView>
template <class Vector>
bool MyFaces<GridView>::xyCollinear(Vector const &a, Vector const &b,
                                    Vector const &c) {
  return isClose((b[0] - a[0]) * (c[1] - a[1]), (b[1] - a[1]) * (c[0] - a[0]));
}

template <class GridView>
template <class Vector>
bool MyFaces<GridView>::xyBoxed(Vector const &v1, Vector const &v2,
                                Vector const &x) {
  auto const minmax0 = std::minmax(v1[0], v2[0]);
  auto const minmax1 = std::minmax(v1[1], v2[1]);

  if (minmax0.first - 1e-14 > x[0] or x[0] > minmax0.second + 1e-14)
    return false;
  if (minmax1.first - 1e-14 > x[1] or x[1] > minmax1.second + 1e-14)
    return false;

  return true;
}

template <class GridView>
template <class Vector>
bool MyFaces<GridView>::xyBetween(Vector const &v1, Vector const &v2,
                                  Vector const &x) {
  return xyCollinear(v1, v2, x) && xyBoxed(v1, v2, x);
}

template <class GridView>
MyFaces<GridView>::MyFaces(GridView const &gridView)
    :
#if DIM == 3
      lower(gridView),
      right(gridView),
      upper(gridView),
      front(gridView),
      back(gridView)
#else
      lower(gridView),
      right(gridView),
      upper(gridView)
#endif
{
  using Grid = typename GridView::Grid;
  assert(isClose(MyGeometry::A[1], 0));
  assert(isClose(MyGeometry::B[1], 0));
  lower.insertFacesByProperty([&](typename GridView::Intersection const &in) {
    return isClose(0, in.geometry().center()[1]);
  });
#if DIM == 3
  front.insertFacesByProperty([&](typename GridView::Intersection const &in) {
    return isClose(-MyGeometry::depth / 2.0, in.geometry().center()[2]);
  });

  back.insertFacesByProperty([&](typename GridView::Intersection const &in) {
    return isClose(MyGeometry::depth / 2.0, in.geometry().center()[2]);
  });
#endif

  upper.insertFacesByProperty([&](typename GridView::Intersection const &in) {
    return xyBetween(MyGeometry::A, MyGeometry::C, in.geometry().center());
  });

  right.insertFacesByProperty([&](typename GridView::Intersection const &in) {
    return xyBetween(MyGeometry::B, MyGeometry::C, in.geometry().center());
  });
}

#include "mygrid_tmpl.cc"
